define('diametric', ['exports'], function (exports) { 'use strict';

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Container = function () {
    function Container(factoryResolver, invocableResolver) {
        _classCallCheck(this, Container);

        this._factoryResolver = factoryResolver;
        this._invocableResolver = invocableResolver;
    }

    Container.prototype.lookup = function lookup(specifier, constructorArgs) {
        var _this = this;

        var factory = this._factoryResolver.factoryFor(specifier);
        var instance = void 0;
        if (!constructorArgs) {
            instance = factory.instance;
            if (instance !== undefined) {
                return instance;
            }
            var constructorInjections = this._factoryResolver.constructorArgumentsFor(specifier);
            constructorArgs = constructorInjections ? constructorInjections.map(function (i) {
                return _this.evaluateInjection(i);
            }) : [];
        }
        instance = factory.create.apply(factory, constructorArgs);
        var initializers = this._factoryResolver.initializersFor(specifier);
        if (initializers) {
            initializers.forEach(function (initializer) {
                var initializerInjections = _this._factoryResolver.initializerArgumentsFor(specifier, initializer.name);
                var initializerArgs = initializerInjections ? initializerInjections.map(function (i) {
                    return _this.evaluateInjection(i);
                }) : [];
                initializer.initialize.apply(initializer, [instance].concat(initializerArgs));
            });
        }
        return instance;
    };

    Container.prototype.teardown = function teardown(specifier, instance) {
        var _this2 = this;

        var i = instance || this.lookup(specifier);
        var destructors = this._factoryResolver.destructorsFor(specifier);
        if (destructors) {
            destructors.forEach(function (destructor) {
                var destructorInjections = _this2._factoryResolver.destructorArgumentsFor(specifier, destructor.name);
                var destructorArgs = destructorInjections ? destructorInjections.map(function (i) {
                    return _this2.evaluateInjection(i);
                }) : [];
                destructor.teardown.apply(destructor, [i].concat(destructorArgs));
            });
        }
    };

    Container.prototype.invoke = function invoke(specifier, args) {
        var _this3 = this;

        var invocable = this._invocableResolver.invocableFor(specifier);
        if (!args) {
            var result = invocable.result;
            if (result !== undefined) {
                return result;
            }
            var injections = this._invocableResolver.invocableArgumentsFor(specifier);
            args = injections ? injections.map(function (i) {
                return _this3.evaluateInjection(i);
            }) : [];
        }
        return invocable.invoke.apply(invocable, args);
    };

    Container.prototype.evaluateInjection = function evaluateInjection(injection) {
        if (typeof injection === 'object') {
            switch (injection.type) {
                case 'static':
                    return this.evaluateStaticInjection(injection);
                case 'creatable':
                    return this.evaluateCreatableInjection(injection);
                case 'property':
                    return this.evaluatePropertyInjection(injection);
                case 'method':
                    return this.evaluateMethodInjection(injection);
                case 'invocable':
                    return this.evaluateInvocableInjection(injection);
                case 'dictionary':
                    return this.evaluateDictionaryInjection(injection);
                case 'object':
                    return this.evaluateObjectInjection(injection);
                case 'array':
                    return this.evaluateArrayInjection(injection);
                default:
                    throw new Error('Unknown injection type: ${injection.type}');
            }
        } else {
            return injection;
        }
    };

    Container.prototype.evaluateStaticInjection = function evaluateStaticInjection(injection) {
        return injection.value;
    };

    Container.prototype.evaluateCreatableInjection = function evaluateCreatableInjection(injection) {
        return this.lookup(injection.source, injection.constructorArgs);
    };

    Container.prototype.evaluatePropertyInjection = function evaluatePropertyInjection(injection) {
        var instance = this.lookup(injection.source, injection.constructorArgs);
        return instance[injection.property];
    };

    Container.prototype.evaluateMethodInjection = function evaluateMethodInjection(injection) {
        var _this4 = this;

        var instance = this.lookup(injection.source, injection.constructorArgs);
        var method = instance[injection.method];
        var args = injection.args ? injection.args.map(function (i) {
            return _this4.evaluateInjection(i);
        }) : [];
        return method.apply(instance, args);
    };

    Container.prototype.evaluateInvocableInjection = function evaluateInvocableInjection(injection) {
        var _this5 = this;

        var args = injection.args ? injection.args.map(function (i) {
            return _this5.evaluateInjection(i);
        }) : undefined;
        return this.invoke(injection.source, args);
    };

    Container.prototype.evaluateDictionaryInjection = function evaluateDictionaryInjection(injection) {
        var _this6 = this;

        var source = injection.value;
        var result = {};
        Object.keys(source).forEach(function (k) {
            result[k] = _this6.evaluateInjection(source[k]);
        });
        return result;
    };

    Container.prototype.evaluateObjectInjection = function evaluateObjectInjection(injection) {
        var keys = injection.keys,
            values = injection.values;

        var result = {};
        for (var i = 0, l = keys.length; i < l; i++) {
            var key = keys[i];
            var value = values[i];
            result[this.evaluateInjection(key)] = this.evaluateInjection(value);
        }
        return result;
    };

    Container.prototype.evaluateArrayInjection = function evaluateArrayInjection(injection) {
        var _this7 = this;

        return injection.values.map(function (i) {
            return _this7.evaluateInjection(i);
        });
    };

    return Container;
}();

function _classCallCheck$1(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var InvocableRegistry = function () {
    function InvocableRegistry(specifierResolver) {
        _classCallCheck$1(this, InvocableRegistry);

        this._specifierResolver = specifierResolver;
        this._invocables = {};
        this._invocableArguments = {};
    }

    InvocableRegistry.prototype.registerInvocable = function registerInvocable(specifier, invocable) {
        var key = this._specifierResolver.specifierKey(specifier);
        this._invocables[key] = invocable;
    };

    InvocableRegistry.prototype.unregisterInvocable = function unregisterInvocable(specifier) {
        var key = this._specifierResolver.specifierKey(specifier);
        delete this._invocables[key];
    };

    InvocableRegistry.prototype.invocableFor = function invocableFor(specifier) {
        var specifiers = this._specifierResolver.matchingSpecifiers(specifier);
        for (var i = 0, l = specifiers.length; i < l; i++) {
            var invocable = this.retrieveInvocable(specifiers[i]);
            if (invocable !== undefined) {
                return invocable;
            }
        }
    };

    InvocableRegistry.prototype.registerInvocableArguments = function registerInvocableArguments(specifier, args) {
        var key = this._specifierResolver.specifierKey(specifier);
        this._invocableArguments[key] = args;
    };

    InvocableRegistry.prototype.unregisterInvocableArguments = function unregisterInvocableArguments(specifier) {
        var key = this._specifierResolver.specifierKey(specifier);
        delete this._invocableArguments[key];
    };

    InvocableRegistry.prototype.invocableArgumentsFor = function invocableArgumentsFor(specifier) {
        var specifiers = this._specifierResolver.matchingSpecifiers(specifier);
        for (var i = 0, l = specifiers.length; i < l; i++) {
            var args = this.retrieveInvocableArguments(specifiers[i]);
            if (args !== undefined) {
                return args;
            }
        }
    };

    InvocableRegistry.prototype.retrieveInvocable = function retrieveInvocable(specifier) {
        var key = this._specifierResolver.specifierKey(specifier);
        return this._invocables[key];
    };

    InvocableRegistry.prototype.retrieveInvocableArguments = function retrieveInvocableArguments(specifier) {
        var key = this._specifierResolver.specifierKey(specifier);
        return this._invocableArguments[key];
    };

    return InvocableRegistry;
}();

function _classCallCheck$2(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var FactoryRegistry = function () {
    function FactoryRegistry(specifierResolver) {
        _classCallCheck$2(this, FactoryRegistry);

        this._specifierResolver = specifierResolver;
        this._factories = {};
        this._initializers = {};
        this._destructors = {};
        this._constructorArguments = {};
        this._initializerArguments = {};
        this._destructorArguments = {};
    }

    FactoryRegistry.prototype.registerFactory = function registerFactory(specifier, factory) {
        var key = this._specifierResolver.specifierKey(specifier);
        this._factories[key] = factory;
    };

    FactoryRegistry.prototype.unregisterFactory = function unregisterFactory(specifier) {
        var key = this._specifierResolver.specifierKey(specifier);
        delete this._factories[key];
    };

    FactoryRegistry.prototype.factoryFor = function factoryFor(specifier) {
        var specifiers = this._specifierResolver.matchingSpecifiers(specifier);
        for (var i = 0, l = specifiers.length; i < l; i++) {
            var factory = this.retrieveFactory(specifiers[i]);
            if (factory !== undefined) {
                return factory;
            }
        }
    };

    FactoryRegistry.prototype.registerInitializer = function registerInitializer(specifier, initializer) {
        var key = this._specifierResolver.specifierKey(specifier);
        this._initializers[key] = this._initializers[key] || {};
        this._initializers[key][initializer.name] = initializer;
    };

    FactoryRegistry.prototype.unregisterInitializer = function unregisterInitializer(specifier, name) {
        var key = this._specifierResolver.specifierKey(specifier);
        if (this._initializers[key]) {
            delete this._initializers[key][name];
        }
    };

    FactoryRegistry.prototype.unregisterInitializers = function unregisterInitializers(specifier) {
        var key = this._specifierResolver.specifierKey(specifier);
        delete this._initializers[key];
    };

    FactoryRegistry.prototype.initializersFor = function initializersFor(specifier) {
        var matches = {};
        var specifiers = this._specifierResolver.matchingSpecifiers(specifier);
        for (var i = specifiers.length - 1; i >= 0; i--) {
            var initializers = this.retrieveInitializers(specifiers[i]);
            if (initializers !== undefined) {
                Object.assign(matches, initializers);
            }
        }
        return Object.values(matches);
    };

    FactoryRegistry.prototype.registerDestructor = function registerDestructor(specifier, destructor) {
        var key = this._specifierResolver.specifierKey(specifier);
        this._destructors[key] = this._destructors[key] || {};
        this._destructors[key][destructor.name] = destructor;
    };

    FactoryRegistry.prototype.unregisterDestructor = function unregisterDestructor(specifier, name) {
        var key = this._specifierResolver.specifierKey(specifier);
        if (this._destructors[key]) {
            delete this._destructors[key][name];
        }
    };

    FactoryRegistry.prototype.unregisterDestructors = function unregisterDestructors(specifier) {
        var key = this._specifierResolver.specifierKey(specifier);
        delete this._destructors[key];
    };

    FactoryRegistry.prototype.destructorsFor = function destructorsFor(specifier) {
        var matches = {};
        var specifiers = this._specifierResolver.matchingSpecifiers(specifier);
        for (var i = specifiers.length - 1; i >= 0; i--) {
            var destructors = this.retrieveDestructors(specifiers[i]);
            if (destructors !== undefined) {
                Object.assign(matches, destructors);
            }
        }
        return Object.values(matches);
    };

    FactoryRegistry.prototype.registerConstructorArguments = function registerConstructorArguments(specifier, args) {
        var key = this._specifierResolver.specifierKey(specifier);
        this._constructorArguments[key] = args;
    };

    FactoryRegistry.prototype.unregisterConstructorArguments = function unregisterConstructorArguments(specifier) {
        var key = this._specifierResolver.specifierKey(specifier);
        delete this._constructorArguments[key];
    };

    FactoryRegistry.prototype.constructorArgumentsFor = function constructorArgumentsFor(specifier) {
        var specifiers = this._specifierResolver.matchingSpecifiers(specifier);
        for (var i = 0, l = specifiers.length; i < l; i++) {
            var args = this.retrieveConstructorArguments(specifiers[i]);
            if (args !== undefined) {
                return args;
            }
        }
    };

    FactoryRegistry.prototype.registerInitializerArguments = function registerInitializerArguments(specifier, name, args) {
        var key = this._specifierResolver.specifierKey(specifier);
        this._initializerArguments[key] = this._initializerArguments[key] || {};
        this._initializerArguments[key][name] = args;
    };

    FactoryRegistry.prototype.unregisterInitializerArguments = function unregisterInitializerArguments(specifier, name) {
        var key = this._specifierResolver.specifierKey(specifier);
        if (this._initializerArguments[key]) {
            if (name) {
                delete this._initializerArguments[key][name];
            } else {
                delete this._initializerArguments[key];
            }
        }
    };

    FactoryRegistry.prototype.initializerArgumentsFor = function initializerArgumentsFor(specifier, name) {
        var specifiers = this._specifierResolver.matchingSpecifiers(specifier);
        for (var i = 0, l = specifiers.length; i < l; i++) {
            var args = this.retrieveInitializerArguments(specifiers[i], name);
            if (args !== undefined) {
                return args;
            }
        }
    };

    FactoryRegistry.prototype.registerDestructorArguments = function registerDestructorArguments(specifier, name, args) {
        var key = this._specifierResolver.specifierKey(specifier);
        this._destructorArguments[key] = this._destructorArguments[key] || {};
        this._destructorArguments[key][name] = args;
    };

    FactoryRegistry.prototype.unregisterDestructorArguments = function unregisterDestructorArguments(specifier, name) {
        var key = this._specifierResolver.specifierKey(specifier);
        if (this._destructorArguments[key]) {
            if (name) {
                delete this._destructorArguments[key][name];
            } else {
                delete this._destructorArguments[key];
            }
        }
    };

    FactoryRegistry.prototype.destructorArgumentsFor = function destructorArgumentsFor(specifier, name) {
        var specifiers = this._specifierResolver.matchingSpecifiers(specifier);
        for (var i = 0, l = specifiers.length; i < l; i++) {
            var args = this.retrieveDestructorArguments(specifiers[i], name);
            if (args !== undefined) {
                return args;
            }
        }
    };

    FactoryRegistry.prototype.retrieveFactory = function retrieveFactory(specifier) {
        var key = this._specifierResolver.specifierKey(specifier);
        return this._factories[key];
    };

    FactoryRegistry.prototype.retrieveInitializers = function retrieveInitializers(specifier) {
        var key = this._specifierResolver.specifierKey(specifier);
        return this._initializers[key];
    };

    FactoryRegistry.prototype.retrieveDestructors = function retrieveDestructors(specifier) {
        var key = this._specifierResolver.specifierKey(specifier);
        return this._destructors[key];
    };

    FactoryRegistry.prototype.retrieveConstructorArguments = function retrieveConstructorArguments(specifier) {
        var key = this._specifierResolver.specifierKey(specifier);
        return this._constructorArguments[key];
    };

    FactoryRegistry.prototype.retrieveInitializerArguments = function retrieveInitializerArguments(specifier, name) {
        var key = this._specifierResolver.specifierKey(specifier);
        if (this._initializerArguments[key]) {
            return this._initializerArguments[key][name];
        }
    };

    FactoryRegistry.prototype.retrieveDestructorArguments = function retrieveDestructorArguments(specifier, name) {
        var key = this._specifierResolver.specifierKey(specifier);
        if (this._destructorArguments[key]) {
            return this._destructorArguments[key][name];
        }
    };

    return FactoryRegistry;
}();

exports.Container = Container;
exports.InvocableRegistry = InvocableRegistry;
exports.FactoryRegistry = FactoryRegistry;

Object.defineProperty(exports, '__esModule', { value: true });

});
